'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.default = lessTokenize;
var singleQuote = '\''.charCodeAt(0);
var doubleQuote = '"'.charCodeAt(0);
var backTick = '`'.charCodeAt(0);
var backslash = '\\'.charCodeAt(0);
var slash = '/'.charCodeAt(0);
var ampersand = '&'.charCodeAt(0);
var newline = '\n'.charCodeAt(0);
var space = ' '.charCodeAt(0);
var feed = '\f'.charCodeAt(0);
var tab = '\t'.charCodeAt(0);
var carriageReturn = '\r'.charCodeAt(0);
var openedParenthesis = '('.charCodeAt(0);
var closedParenthesis = ')'.charCodeAt(0);
var openedСurlyBracket = '{'.charCodeAt(0);
var closedCurlyBracket = '}'.charCodeAt(0);
var semicolon = ';'.charCodeAt(0);
var asterisk = '*'.charCodeAt(0);
var colon = ':'.charCodeAt(0);
var comma = ','.charCodeAt(0);
var atRule = '@'.charCodeAt(0);
var tilde = '~'.charCodeAt(0);
var dot = '.'.charCodeAt(0);
var hash = '#'.charCodeAt(0);
var atEndPattern = /[ \n\t\r\f\{\(\)'"`\\;/]/g;
var wordEndPattern = /[ \n\t\r\f\(\)\{\}:;@!'"`\\#]|\/(?=\*)/g;
var badBracketPattern = /.[\\\/\("'\n]/;
var variablePattern = /^@[^:\(\{]+:/;
var nestedExtendWord = '&:extend';
var nestedExtendWordLength = nestedExtendWord.length;

function isMixinName(str) {
    var firstSymbolCode = str ? str[0].charCodeAt(0) : null;

    return firstSymbolCode === dot || firstSymbolCode === hash;
}

function findEndOfExpression(css, length, j) {
    var openedParenthesisBlocks = 0;
    var openedCurlyBlocks = 0;

    while (j < length) {
        var symbolCode = css[j].charCodeAt(0);

        //find the on of escaped expression
        if (!openedParenthesisBlocks && !openedCurlyBlocks && (symbolCode === semicolon || symbolCode === closedCurlyBracket)) {
            return j - 1;
        }

        switch (symbolCode) {
            case openedСurlyBracket:
                {
                    openedCurlyBlocks++;
                    break;
                }

            case closedCurlyBracket:
                {
                    openedCurlyBlocks--;
                    break;
                }

            case openedParenthesis:
                {
                    openedParenthesisBlocks++;
                    break;
                }

            case closedParenthesis:
                {
                    openedParenthesisBlocks--;
                    break;
                }
        }

        j++;
    }

    return -1;
}

function lessTokenize(input) {
    var tokens = [];
    var css = input.css.valueOf();
    var length = css.length;

    var offset = -1;
    var line = 1;
    var pos = 0;

    function unclosed(what) {
        throw input.error('Unclosed ' + what, line, pos - offset);
    }

    while (pos < length) {
        var symbolCode = css.charCodeAt(pos);
        var symbol = css[pos];
        var nextSymbolCode = undefined;
        var nextPos = undefined;
        var escaped = undefined;
        var lines = undefined;
        var lastLine = undefined;
        var cssPart = undefined;
        var escape = undefined;
        var nextLine = undefined;
        var nextOffset = undefined;
        var escapePos = undefined;
        var token = undefined;

        if (symbolCode === newline) {
            offset = pos;
            line += 1;
        }

        switch (symbolCode) {
            case newline:
            case space:
            case tab:
            case carriageReturn:
            case feed:
                {
                    nextPos = pos;

                    //collect all neighbour space symbols
                    do {
                        nextPos += 1;
                        symbolCode = css.charCodeAt(nextPos);

                        if (symbolCode === newline) {
                            offset = nextPos;
                            line += 1;
                        }
                    } while (symbolCode === space || symbolCode === newline || symbolCode === tab || symbolCode === carriageReturn || symbolCode === feed);

                    tokens.push(['space', css.slice(pos, nextPos)]);
                    pos = nextPos - 1;
                    break;
                }

            case comma:
                {
                    tokens.push(['word', symbol, line, pos - offset, line, pos - offset + 1]);

                    break;
                }

            case colon:
            case semicolon:
            case openedСurlyBracket:
            case closedCurlyBracket:
            case closedParenthesis:
                {
                    tokens.push([symbol, symbol, line, pos - offset]);

                    break;
                }

            case openedParenthesis:
                {
                    nextSymbolCode = css.charCodeAt(pos + 1);
                    var tokensCount = tokens.length;
                    var prevTokenCssPart = tokensCount ? tokens[tokensCount - 1][1] : '';

                    if (prevTokenCssPart === 'url' && nextSymbolCode !== singleQuote && nextSymbolCode !== doubleQuote && nextSymbolCode !== space && nextSymbolCode !== newline && nextSymbolCode !== tab && nextSymbolCode !== feed && nextSymbolCode !== carriageReturn) {
                        nextPos = pos;

                        do {
                            escaped = false;
                            nextPos = css.indexOf(')', nextPos + 1);
                            if (nextPos === -1) {
                                unclosed('bracket');
                            }

                            escapePos = nextPos;
                            while (css.charCodeAt(escapePos - 1) === backslash) {
                                escapePos -= 1;
                                escaped = !escaped;
                            }
                        } while (escaped);

                        tokens.push(['brackets', css.slice(pos, nextPos + 1), line, pos - offset, line, nextPos - offset]);

                        pos = nextPos;
                    } else {
                        nextPos = css.indexOf(')', pos + 1);
                        cssPart = css.slice(pos, nextPos + 1);

                        if (nextPos === -1 || badBracketPattern.test(cssPart)) {
                            tokens.push(['(', '(', line, pos - offset]);
                        } else {
                            tokens.push(['brackets', cssPart, line, pos - offset, line, nextPos - offset]);

                            pos = nextPos;
                        }
                    }

                    break;
                }

            case singleQuote:
            case doubleQuote:
                {
                    nextPos = pos;
                    do {
                        escaped = false;
                        nextPos = css.indexOf(symbol, nextPos + 1);
                        if (nextPos === -1) {
                            unclosed('quote');
                        }

                        escapePos = nextPos;
                        while (css.charCodeAt(escapePos - 1) === backslash) {
                            escapePos -= 1;
                            escaped = !escaped;
                        }
                    } while (escaped);

                    tokens.push(['string', css.slice(pos, nextPos + 1), line, pos - offset, line, nextPos - offset]);

                    pos = nextPos;
                    break;
                }

            case atRule:
                {
                    // interpolation: @{...
                    if (css.charCodeAt(pos + 1) === openedСurlyBracket) {
                        nextPos = css.indexOf('}', pos + 2);
                        if (nextPos === -1) {
                            unclosed('interpolation');
                        }

                        cssPart = css.slice(pos, nextPos + 1);
                        lines = cssPart.split('\n');
                        lastLine = lines.length - 1;

                        if (lastLine > 0) {
                            nextLine = line + lastLine;
                            nextOffset = nextPos - lines[lastLine].length;
                        } else {
                            nextLine = line;
                            nextOffset = offset;
                        }

                        tokens.push(['word', cssPart, line, pos - offset, nextLine, nextPos - nextOffset]);

                        offset = nextOffset;
                        line = nextLine;
                    } else {
                        atEndPattern.lastIndex = pos + 1;
                        atEndPattern.test(css);
                        if (atEndPattern.lastIndex === 0) {
                            nextPos = css.length - 1;
                        } else {
                            nextPos = atEndPattern.lastIndex - 2;
                        }

                        cssPart = css.slice(pos, nextPos + 1);
                        token = 'at-word';

                        //check if it's a variable
                        if (variablePattern.test(cssPart)) {
                            wordEndPattern.lastIndex = pos + 1;
                            wordEndPattern.test(css);
                            if (wordEndPattern.lastIndex === 0) {
                                nextPos = css.length - 1;
                            } else {
                                nextPos = wordEndPattern.lastIndex - 2;
                            }

                            cssPart = css.slice(pos, nextPos + 1);
                            token = 'word';
                        }

                        tokens.push([token, cssPart, line, pos - offset, line, nextPos - offset]);
                    }

                    pos = nextPos;
                    break;
                }

            case backslash:
                {
                    nextPos = pos;
                    escape = true;

                    while (css.charCodeAt(nextPos + 1) === backslash) {
                        nextPos += 1;
                        escape = !escape;
                    }

                    symbolCode = css.charCodeAt(nextPos + 1);
                    if (escape && symbolCode !== slash && symbolCode !== space && symbolCode !== nextLine && symbolCode !== tab && symbolCode !== carriageReturn && symbolCode !== feed) {
                        nextPos += 1;
                    }

                    tokens.push(['word', css.slice(pos, nextPos + 1), line, pos - offset, line, nextPos - offset]);

                    pos = nextPos;
                    break;
                }

            default:
                {
                    //it's a nested &:extend
                    if (symbolCode === ampersand && css.slice(pos, pos + nestedExtendWordLength) === nestedExtendWord) {
                        var end = findEndOfExpression(css, length, pos + nestedExtendWordLength);

                        if (end !== -1) {
                            tokens.push(['nested-extend', css.slice(pos, end + 1), line, pos - offset, nextLine, end - nextOffset]);

                            pos = end;
                        }

                        break;
                    }

                    nextSymbolCode = css.charCodeAt(pos + 1);

                    if (symbolCode === slash && nextSymbolCode === asterisk) {
                        nextPos = css.indexOf('*/', pos + 2) + 1;
                        if (nextPos === 0) {
                            unclosed('comment');
                        }

                        cssPart = css.slice(pos, nextPos + 1);
                        lines = cssPart.split('\n');
                        lastLine = lines.length - 1;

                        if (lastLine > 0) {
                            nextLine = line + lastLine;
                            nextOffset = nextPos - lines[lastLine].length;
                        } else {
                            nextLine = line;
                            nextOffset = offset;
                        }

                        tokens.push(['comment', cssPart, line, pos - offset, nextLine, nextPos - nextOffset]);

                        offset = nextOffset;
                        line = nextLine;
                        pos = nextPos;
                    } else if (symbolCode === slash && nextSymbolCode === slash) {
                        nextPos = css.indexOf('\n', pos + 2) - 1;
                        if (nextPos === -2) {
                            nextPos = css.length - 1;
                        }

                        tokens.push(['comment', css.slice(pos, nextPos + 1), line, pos - offset, line, nextPos - offset, 'inline']);

                        pos = nextPos;
                    } else {
                        var isEscaping = false;

                        if (symbolCode === tilde) {
                            for (var j = pos + 1; j < length; j++) {
                                symbolCode = css[j].charCodeAt(0);

                                //start of escaped expression
                                if (symbolCode === singleQuote || symbolCode === doubleQuote || symbolCode === backTick) {
                                    var end = findEndOfExpression(css, length, j + 1);
                                    if (end !== -1) {
                                        isEscaping = true;
                                        nextPos = end;
                                    }

                                    break;
                                }
                            }
                        }

                        if (!isEscaping) {
                            wordEndPattern.lastIndex = pos + 1;
                            wordEndPattern.test(css);

                            if (wordEndPattern.lastIndex === 0) {
                                nextPos = css.length - 1;
                            } else {
                                nextPos = wordEndPattern.lastIndex - 2;
                            }
                        }

                        cssPart = css.slice(pos, nextPos + 1);

                        var tokensCount = tokens.push(['word', cssPart, line, pos - offset, line, nextPos - offset]);

                        /**
                         * - mark nested mixin by custom token 'nested-mixin'
                         * - skip variables that start from dot (.5s) or from hash (#f3f3f3)
                         */
                        if (isMixinName(cssPart) && (tokens[tokensCount - 2] || [])[1] !== ':' && (tokens[tokensCount - 3] || [])[1] !== ':') {
                            var openedParenthesisBlocks = 0;
                            var openedCurlyBlocks = 0;
                            var endLine = line;
                            var endOffset = offset;

                            for (var j = pos + 1; j < length; j++) {
                                symbolCode = css.charCodeAt(j);
                                endOffset++;

                                if (!openedParenthesisBlocks && !openedCurlyBlocks && (symbolCode === semicolon || symbolCode === closedCurlyBracket)) {
                                    nextPos = j;

                                    //not including include '{' symbol
                                    if (symbolCode === closedCurlyBracket) {
                                        nextPos--;
                                    }

                                    cssPart = css.slice(pos, nextPos + 1);
                                    tokens.pop();
                                    tokens.push(['nested-mixin', cssPart, line, pos - offset, endLine, nextPos - endOffset]);

                                    break;
                                }

                                switch (symbolCode) {
                                    case newline:
                                        {
                                            endLine++;
                                            endOffset = j;
                                            break;
                                        }

                                    case openedСurlyBracket:
                                        {
                                            //it's a CSS block -> exit from the loop
                                            if (!openedCurlyBlocks && !openedParenthesisBlocks) {
                                                j = length;
                                            }

                                            openedCurlyBlocks++;
                                            break;
                                        }

                                    case closedCurlyBracket:
                                        {
                                            openedCurlyBlocks--;
                                            break;
                                        }

                                    case openedParenthesis:
                                        {
                                            openedParenthesisBlocks++;
                                            break;
                                        }

                                    case closedParenthesis:
                                        {
                                            openedParenthesisBlocks--;
                                            break;
                                        }
                                }
                            }
                        }

                        pos = nextPos;
                    }

                    break;
                }
        }

        pos++;
    }

    return tokens;
}